# coding=utf-8
# Generate a set of contracts

# GenerateContracts is the main function. The rest are mostly
# support functions that handles certain properties.

import numpy as np
from scipy import stats
import datetime as dt
import random as rd
from scipy.special import factorial
import matplotlib.pyplot as plt

# My programs
from ContractRigObjects import Contract
from ContractRigObjects import getNewTime

#-------- Exogenous parameters --------#


def getOilPrice(T, p_oil_price):
    """
        Description: generate an array of oil prices 
        per year from a normal distribution.

        Input: time period (years) and parameters N(mu,sigma) 
        for oil price.
        Output: yearly average oil price - Tx1 array
    """

    oil_p = np.zeros([T, 1])

    mu_oil = p_oil_price[0]
    sigma_oil = p_oil_price[1]

    # Other oil prices - depend on previous price
    for i in range(len(oil_p)):
        while oil_p[i] <= 0:
            oil_p[i] = np.random.normal(mu_oil, sigma_oil)

    return oil_p

#-------- Contract Parameter --------#


# def getNumContracts(oil_price, p_num_contracts):
def getNumContracts(oil_price, fleet_size, p_num_contracts):
    """ 
    Description: generate a set of N contracts per year.
    Input: oil price, number of rigs in fleet (V) and number of
    contracts per rig (N/V rate).
    Output: number of contracts for year t.

    """
    rate = p_num_contracts[0, 1]
    for i in p_num_contracts:
        if i[0] <= oil_price:
            rate = i[1]

    mu_N = fleet_size * rate
    sigma_N = 0.1*mu_N

    N = -1
    while N <= 0:  # Require at least one contract
        N = np.random.normal(mu_N, sigma_N)
        N = int(round(N))

    return N


def getMarket(p_market):
    """ 
        Return which market contract belongs to.
        Input: probability array
        Output: market selection
    """
    xk = np.arange(len(p_market))
    market = ['B', 'GOM', 'NS', 'WA']
    # custom discrete probability distribution
    custm = stats.rv_discrete(name='custm', values=(xk, p_market))
    i = custm.rvs()
    return market[i]

# ---------- Start Requirements -----------------


def reqEnvironment(market):
    """ 
        Harsh environment = 1, Benign environment = 0
    """
    if market == 'NS':
        he_be = 1
    else:
        he_be = 0
    return he_be


def reqWaterDepth(market, p_depths):
    """
        Input: market indicator, prob. depth array
        Output: depth of contract

        Water depth => shallow (<125m), mid (<1500m) and deep (>1500m)
    """
    xk = np.arange(len(p_depths[0, 1:]))

    for a in p_depths[1:]:
        if market in a[0]:
            pk = a[1:]
            break
    custm = stats.rv_discrete(name='custm', values=(xk, pk))
    i = custm.rvs()
    water_depth = p_depths[0, i+1]
    return water_depth


def reqContractScope(p_scope):
    """ 
        Development contract = 0
        Exploration contract (wildcat, leting) = 1, 
        Appraisal contract (avgrensning) = 2,

        # 74 % development wells (utvinning) - Oljedirektoratet
        # 17 % exploration contract (undersøkelse)
        # 9 % appraisal contract (avgrensning)
    """
    xk = np.arange(len(p_scope))
    custm = stats.rv_discrete(name='custm', values=(xk, p_scope))
    scope = custm.rvs()

    return scope
# ---------- End Requirements -------------------


def getStartTimeContract(year):
    """
        Input:
        year - index counting distance from year zero.
        time_commence - number of months from sign date to 
        start date of contract.
        Output: start date of contract
    """

    t_signed = dt.date(2020 + year, rd.randint(1, 12),
                       1)

    return t_signed


def getDuration(oil_price, betas):
    """
        Get duration of contract in months
    """

    # expected duration
    dur_mu = betas[0] + betas[1]*oil_price  # duration in days
    dur_mu = dur_mu/30.4  # in months (avg. 30.4 days)
    dur_sigma = 0.8*dur_mu  # 80% stddev

    duration = 1
    while duration <= 1:
        duration = np.random.normal(dur_mu, dur_sigma)

    duration = int(duration)  # rounding down

    return duration


def getRevenue(oil_price, p_revenue, rev_prems, market, depth,
               duration, scope):
    """
        Get revenue of contract in USD/day. Compute base dayrate
        based on water depth. Add premiums for high specification
        contracts.
    """

    if depth == 50:  # shallow water => JU equation
        # log_revenue = p_revenue[0, 0] + \
        #     p_revenue[0, 1]*np.log(oil_price)
        revenue = p_revenue[0, 0]+p_revenue[0, 1]*oil_price
    else:                   # Semi or DS
        # log_revenue = p_revenue[1, 0] + \
        #     p_revenue[1, 1] * np.log(oil_price)
        revenue = p_revenue[1, 0]+p_revenue[1, 1]*oil_price

    # revenue = np.exp(log_revenue)  # standard pay/day
    # print "Standard dayrate: %.2f USD/day" % (revenue)

    # Add premiums
    premium = 0
    if market == 'NS':  # => Harsh env.
        premium = premium + rev_prems[0]*revenue  # add 15 %
    if depth == 2000:  # midwater premium included in floater category
        premium = premium + rev_prems[1]*revenue  # add 35 %
    if duration >= 18:  # longer than 1.5 year contract
        premium = premium + rev_prems[2]*revenue
    if scope == 2:  # appraisal contract
        premium = premium + rev_prems[3]*revenue

    rev_mu = revenue + premium  # exp.value
    rev_sigma = 0.2*rev_mu  # std.dev of 20 %

    revenue = 0
    while revenue <= 0:
        revenue = np.random.normal(rev_mu, rev_sigma)

    revenue = np.around(revenue, 0)  # round to no decimals

    return revenue

#-------- Generate Contracts (main) --------#


def generateContracts(T, p_oil_price, fleet_size, p_num_contracts,
                      p_market, p_depths, p_scope, p_duration,
                      p_revenue, rev_prems):
    """
        Generate a set of contracts with inherent attributes.

        Input:
        T - simulation period (years)
        p_oil_price - expected value and std.dev. of oil price
        fleet_size - number of units in fleet, V
        p_num_contracts - number of contracts per rig
        for different oil prices (N/V ratio)
        p_market - probability distribution of markets
        p_depths - probability distribution of water depths
        p_scope - probability distribution of contract scopes
        betas_durations - coefficients for computation of duration
        p_revenues - coefficients for computation of revenue
        rev_prems - percentage premiums for contract specification

        Output: list of contracts
    """

    # Compute exogenous parameters
    oil_price = getOilPrice(T, p_oil_price)  # return array of T entries
    # Average oil price over ten years and std.dev.
    it_info = np.array([np.mean(oil_price), np.std(oil_price), ])

    # generate N contracts per year,
    listofContracts = np.empty(0)
    count = 1
    for year in range(T):
        N = getNumContracts(oil_price[year], fleet_size, p_num_contracts)

        for i in range(N):
            market = getMarket(p_market)
            he_be = reqEnvironment(market)
            water_depth = reqWaterDepth(market, p_depths)
            contract_scope = reqContractScope(p_scope)

            # Time
            start_time = getStartTimeContract(year)
            # avg time between signing and commencement
            duration = getDuration(oil_price[year], p_duration)
            end_time = getNewTime(start_time, duration)

            revenue_day = getRevenue(oil_price[year], p_revenue, rev_prems,
                                     market, water_depth, duration,
                                     contract_scope)
            # USD/day
            revenue_month = revenue_day*30.4/10**6  # MUSD/month

            e = Contract(count, market, start_time, end_time,
                         duration, revenue_month, he_be,
                         water_depth, contract_scope)
            listofContracts = np.append(listofContracts, e)
            count = count + 1

    return listofContracts, it_info
