import numpy as np
import datetime as dt

from LongestPath import Arc


class Contract:
    numContracts = 0

    def __init__(self, id, market, t_start, t_end, duration, revenue_month,
                 he_be, water_depth, scope):
        self.id = id
        self.market = market
        self.t_start = t_start  # date object (year, month, day)
        self.t_end = t_end
        self.duration = duration
        self.revenue_month = revenue_month
        self.revenue = duration*revenue_month
        # Specifications
        self.he_be = he_be
        self.water_depth = water_depth
        self.scope = scope  # appraisal = 2, exploration = 1, development = 0

        Contract.numContracts += 1

    def getMarket(self):
        """
            Method: can be called like an attribute.
            String representation of markets.
        """
        if self.market == 'B':
            mkt = 'Brazil'
        elif self.market == 'GOM':
            mkt = 'Gulf of Mexico'
        elif self.market == 'NS':
            mkt = 'North Sea'
        else:
            mkt = 'West Africa'
        return mkt

    def getDates(self):
        """
            String representation of start/end dates.
        """
        if self.t_start.month < 10:
            s_month = '0%s' % (str(self.t_start.month))
        else:
            s_month = str(self.t_start.month)
        s_year = str(self.t_start.year)
        if self.t_end.month < 10:
            e_month = '0%s' % (str(self.t_end.month))
        else:
            e_month = str(self.t_end.month)
        e_year = str(self.t_end.year-2000)
        return s_month, s_year, e_month, e_year

    def getEnv(self):
        if self.he_be == 0:
            hebe = 'BE'
        else:
            hebe = 'HE'
        return hebe

    def plotSchedule(self):
        pass

    def __repr__(self):
        s_month, s_year, e_month, e_year = self.getDates()
        hebe = self.getEnv()
        return 'Contract %s: period %s.%s-%s.%s, %sM(USD) - %s, %s, %sm. \n' \
            % (str(self.id), s_month, s_year, e_month, e_year,
               str(self.revenue), self.market, hebe, str(self.water_depth))


class Rig:
    numRigs = 0

    def __init__(self, id, name, rig_type, t_available, cost_purchase,
                 opex, he_be, water_depth, age, speed):
        self.id = id
        self.name = name
        self.type = rig_type
        self.t_available = t_available
        self.cost_purchase = cost_purchase
        self.opex = opex
        # Capacities
        self.he_be = he_be
        self.water_depth = water_depth
        self.age = age
        self.speed = speed

        # Results
        self.profit = 0
        self.servicingcontracts = list()
        self.rigcontracts = list()
        self.rig_months = 0
        self.markets = [0, 0, 0, 0]
        self.market_transitions = 0
        self.contract_scope = [0, 0, 0]
        self.depths = [0, 0, 0]

        Rig.numRigs += 1

    # Methods - properties of contracts awarded, results
    def getContractsServed(self, contracts):
        """
            Get contract objects for all contracts that are being served.
        """
        for i in self.servicingcontracts:
            for j in contracts:
                if i == j.id:
                    self.rigcontracts.append(j)

    def getSumDurations(self):
        """
            Get total time (months) in active work.
        """

        for i in self.rigcontracts:
            self.rig_months = self.rig_months + i.duration

    def getSumMarkets(self):
        """
            Get (1) number of contracts in each market and (2) number of
            transitions between markets.
        """
        # (1)
        for i in self.rigcontracts:
            if i.market == 'B':
                self.markets[0] = self.markets[0] + 1
            elif i.market == 'GOM':
                self.markets[1] = self.markets[1] + 1
            elif i.market == 'NS':
                self.markets[2] = self.markets[2] + 1
            else:
                self.markets[3] = self.markets[3] + 1

        # (2)
        for i in range(1, len(self.rigcontracts)):
            if self.rigcontracts[i].market != self.rigcontracts[i-1].market:
                self.market_transitions = self.market_transitions + 1

    def getDepths(self):
        for i in self.rigcontracts:
            if i.water_depth == 50:
                self.depths[0] = self.depths[0] + 1
            elif i.water_depth == 1500:
                self.depths[1] = self.depths[1] + 1
            else:
                self.depths[2] = self.depths[2] + 1

    def getScope(self):
        for i in self.rigcontracts:
            if i.scope == 0:  # dev
                self.contract_scope[0] = self.contract_scope[0] + 1
            elif i.scope == 1:  # exp
                self.contract_scope[1] = self.contract_scope[1] + 1
            else:  # appr
                self.contract_scope[2] = self.contract_scope[2] + 1

    # Methods for self representation
    def getAvailability(self):

        if self.t_available.month < 10:
            s_month = '0%s' % (str(self.t_available.month))
        else:
            s_month = str(self.t_available.month)
        s_year = str(self.t_available.year)

        return '%s.%s' % (s_month, s_year)

    def getEnv(self):
        if self.he_be == 0:
            hebe = 'BE'
        else:
            hebe = 'HE'
        return hebe

    def __repr__(self):
        t_avail = self.getAvailability()
        hebe = self.getEnv()
        # TODO: Update string representation for rig
        return 'Rig %s: Available %s, cost %sM(USD), - %s, %sm. \n' \
               % (str(self.id), t_avail, str(self.cost_purchase), hebe,
                  str(self.water_depth))

# ---------------- Support functions --------------------------


def dot(x, y):
    """
        Description: dot product of lists.

        Input: dot product LISTS
        Output: dot product of lists
    """
    return sum(x_i * y_i for x_i, y_i in zip(x, y))


def getNewTime(start_time, duration):
    """
        Return new datetime object that progresses a certain number of months.

        Input: datetime object, duration_in_months (int type)
        Output: datetime_object with added months
    """

    month = duration % 12  # get remainder = number of months
    year = (duration-month) / 12  # get number of years

    end_month = (start_time.month + month) % 12  # in case of new year
    if end_month == 0:  # 12 % 12 = 0, we are in 12th month
        end_month = 12
    # integer operation returns 1 if answer >= 1
    new_year = (start_time.month + month - 1) / \
        12  # if start_time.month + month = 12
    # then we are in the 12th month, not a new year -> add (-1)
    end_time = dt.date(start_time.year + year + new_year, end_month, 1)
    return end_time

# ---------- Compatible rigs and feasible arcs ----------------


def compatibleRigs(rig, contracts):
    """
        Description: Create list of nodes compatible with a specific rig.
        A series of conditions have to be satisfied for rig to be compatible:
        1. Rig availabile at contract start time
        2. Rig satisfy environment condition - HE/BE
        3. Rig capable of operating at required water depth
        4. ...


        Input: Rig Object and list of Contract Objects
        Output: List of contracts compatible with current rig design [0, n1, n2,
        ..., end_node]
    """
    nodes_list = [0]  # list with initial node 0
    # Check if list is empty.
    if len(contracts) == 0:
        # length zero defined as no contracts available
        contracts_in_list = np.empty(0)
    else:

        for i in range(len(contracts)):
            # Check 1: Contract termination times - rigs under contract?
            if contracts[i].t_start >= rig.t_available:  # contract starts
                # after termination
                # nodes_list.append(contracts[i].id)

                # REQUIREMENTS
                # Req. 1 weather condition
                if contracts[i].he_be <= rig.he_be:
                    # Req. 2 water depth
                    if contracts[i].water_depth <= rig.water_depth:
                        # Req. 3 scope - JU cannot serve other than development contracts
                        # development contract, all avail., not jackup all availablr
                        if contracts[i].scope == 0 or rig.type != 'Jackup':
                            nodes_list.append(contracts[i].id)

        nodes_list.append(contracts[i].id + 1)  # add end node

        contracts_in_list = getContractsInList(contracts, nodes_list)

    return nodes_list, contracts_in_list


def getContractsInList(contracts, nodes_list):
    """
        Return contracts i
        Input: list of all contracts
        Output: list of feasible contracts
    """
    contracts_in_list = np.empty(0)
    for i in contracts:
        if i.id in nodes_list:
            contracts_in_list = np.append(contracts_in_list, i)
    return contracts_in_list


def getDistanceTransit(contract_to, contract_from, market_distance):
    """
        Return distance between two markets (read market_distance array)
    """
    if contract_to.market == contract_from.market:
        distance_market = 0
    else:
        # Different markets - find in array the distance (nm) between markets
        for i in market_distance:
            if contract_to.market in i and contract_from.market in i:
                distance_market = i[2]
                break  # stop searchung when answer is found
    return distance_market


def feasibleArcs(nodes, f_contracts, rig, market_distance):
    """
        Description: creates a list of feasible Arc objects. Arc costs are 
        revenue from contracts minus operating cost of rig. Design rigs have
        a purchase cost, which is added to end nodes.

        Input: list of feasible nodes, Node object, feasible Contract objects
        and rig object.

        Output: list of Arc objects (startNode, endNode, cost). 
        To be inputed network. Cost of rig purchase is added to 
        final node (i, end, rigCost).
    """

    # print "Feasible arcs for %s. \n" % (str(rig))
    # print "Contract options are "
    # for i in nodes[1:-1]:
    #     print "no. %s, " % (str(i.id))

    # Empty list of arcs
    numarcs = list()

    # (I) Create arcs from start node to all contract nodes
    # Profit = contract revenue - rig opex
    # Assume no transit for first contract
    numarcs = [Arc(nodes[0], nodes[j], f_contracts[j-1].revenue -
                   rig.opex*f_contracts[j-1].duration) for j in range(1, len(nodes)-1)]

    for a in range(1, len(nodes)-1):
        # (II) Create arcs from all contract nodes to end node
        # Cost of rig purchase (capex) added to end node
        numarcs.append(Arc(nodes[a], nodes[-1], -rig.cost_purchase))

        # (III) Create arcs between all contracts that can be sequentially served
        for b in range(1, len(nodes)-1):
            if a is not b:  # no arcs to itself

                # Compute distance betw. markets
                distance_market = getDistanceTransit(
                    f_contracts[a-1], f_contracts[b-1], market_distance)

                # print "Distance from contract %s (%s) to %s (%s) is %s nm.\n" % \
                #     (str(f_contracts[b-1].id), str(f_contracts[b-1].market),
                #      str(f_contracts[a-1].id), str(f_contracts[a-1].market),
                #      str(distance_market))
                # Compute time of transit
                # speed (nm/hr) x 24 hrs x 30 days = speed (nm/month)
                time_transit = distance_market / (rig.speed*720)

                # print "Transit time for rig %s is %s months.\n" % (
                #     str(rig.id), str(time_transit))

                # Compute datetime object that accounts for contract end, plus
                # transit time to next contract
                earliest_start = getNewTime(
                    f_contracts[b-1].t_end, int(time_transit))
                # print f_contracts[b-1].t_end, time_transit, earliest_start

                # Check if contract a starts after b ends + transit time
                if f_contracts[a-1].t_start >= earliest_start:

                    # Cost of transit (MUSD)
                    cost_transit = rig.opex * time_transit
                    # OPEX (MUSD)
                    opex_cost = rig.opex*f_contracts[a-1].duration

                    numarcs.append(
                        Arc(nodes[b], nodes[a], f_contracts[a-1].revenue - opex_cost -
                            cost_transit))
    return numarcs
